
"""
Leases management : 
download from r2labapi service,
creation update deletions,
and minimal verifications
"""

import os
import pwd
import time
from datetime import datetime
import json
import uuid
import traceback

from .logger import logger
from .config import Config
from .plcapiproxy import PlcApiProxy

debug = False
debug = True


class Lease:
    """
    a single lease entry
    """

    wire_timeformat = "%Y-%m-%dT%H:%M:%S%Z"

    def __init__(self, plc_lease):
        r = plc_lease
        try:
            self.leases_hostname = Config().value('plcapi', 'leases_hostname')
            self.owner = r['name']
            self.lease_id = r['lease_id']
            self.subjects = [r['hostname']]
            self.ifrom, self.iuntil = r['t_from'], r['t_until']
            self.broken = False
            # this is only to get __repr__ as short as possible

        except Exception as e:
            self.broken = "lease broken b/c of exception {}".format(e)

        if not self.subjects:
            self.broken = "(no component)"

    # show a 2-chars prefix
    # one for rights (computed by caller - not shown if not provided)
    # one for situation wrt now: < for past, = for current, > for future
    def __repr__(self, rights_char=None):
        rights_char = 'B' if self.broken \
                      else rights_char if rights_char \
            else '?'
        now = time.time()
        if self.iuntil < now:
            time_char = '<'
        elif self.ifrom < now:
            time_char = '='
        else:
            time_char = '>'
        time_message = 'from {} until {}'.format(
            self.human(self.ifrom, show_timezone=False),
            self.human(self.iuntil, show_date=False))
        # usual case is that self.subjects == [leases_hostname]
        if len(self.subjects) == 1 \
           and self.subjects[0] == self.leases_hostname:
            scope = ""
        else:
            scope = " -> {}".format(" & ".join(self.subjects))
        return "{}{} {} {} {}".format(rights_char, time_char,
                                      time_message, self.owner, scope)

    def sort_key(self):
        """
        Sort on start time
        """
        return self.ifrom

    @staticmethod
    def human(epoch, show_date=True, show_timezone=True):
        format = ""
        """
        human-readable format
        """
        if show_date:
            format += "%m-%d @ "
        format += "%H:%M"
        if show_timezone:
            format += " %Z"
        return time.strftime(format, time.localtime(epoch))


    def booked_now_by(self, hostname, login):
        """
        tells if this lease is held right now, by that login (if not None)
        or by anyone (if login is None)
        """
        return self.booked_at_by(hostname, time.time(), login)

    def booked_at_by(self, hostname, instant, login):
        """
        tells if this lease is held at that time, by that login (if not None)
        or by anyone (if login is None)
        """
        if self.broken:
            if debug:
                logger.info("ignoring broken lease {}".format(self))
            return False
        if not self.ifrom <= instant <= self.iuntil:
            if debug:
                logger.info("{} : wrong timerange".format(self))
            return False
        if hostname not in self.subjects:
            if debug:
                logger.info("{} not among subjects {}"
                            .format(hostname, self.subjects))
            return False
        if login is not None and not self.owner == login:
            if debug:
                logger.info(
                    "login {} is not owner - actual owner is {}"
                    .format(login, self.owner))
            return False
        return self


class Leases:
    """
    A list of leases as downloaded from the API
    """
    # the details of the plcapi_proxy instance where to look for leases

    def __init__(self, message_bus):
        self.message_bus = message_bus
        # don't use os.getlogin() as this gives root if under su
        self.login = pwd.getpwuid(os.getuid())[0]
        # the hostname of the plcapi node that we attach leases to
        self.leases_hostname = Config().value('plcapi', 'leases_hostname')
        plcapi_url = Config().value('plcapi', 'url')
        self.plcapi_proxy = PlcApiProxy(plcapi_url)
        # computed later
        # a list of Lease objects
        self.leases = None
        # the result of GetLeases - essentially as-is
        self.plc_leases = None
        # xxx this is still used by monitor
        # should be cleaned up
        self.resources = None

    def __repr__(self):
        if self.leases is None:
            return "<Leases from {} - **(UNFETCHED)**>"\
                .format(self.plcapi_proxy)
        else:
            return "<Leases from {} - {} lease(s)>"\
                .format(self.plcapi_proxy, len(self.leases))

    async def feedback(self, field, msg):
        """
        send feedback, for displaying or monitoring
        """
        await self.message_bus.put({field: msg})

    def has_special_privileges(self):
        """
        check for being run as root
        """
        privileged = Config().value('accounts', 'privileged')
        names = privileged.split(',')
        return self.login in names

    async def booked_now_by_me(self, *, root_allowed=True):
        """
        fetch leases and return a bool that says if current login has a lease
        if root_allowed is True, then this function will return True for root user
           no matter what else
        if root_allowed is false, root goes through the usual process, which most of 
           th time results in this function answering False
        """
        return await self.booked_now_by(login=self.login, root_allowed=root_allowed)

    async def booked_now_by(self, login, *, root_allowed=True):
        if root_allowed and self.has_special_privileges():
            return True
        try:
            await self.fetch_all()
            return self._booked_now_by_login(login)
        except Exception as e:
            await self.feedback('info',
                                "Could not fetch leases : {}".format(e))
            return False

    async def booked_now_by_anyone(self):
        """
        fetch leases and return a bool that says if a lease is currently valid
        """
        try:
            await self.fetch_all()
            return self._booked_now_by_anyone()
        except Exception as e:
            await self.feedback('info', "Could not fetch leases : {}"
                                .format(e))
            return False

    # the following 2 methods assume the leases have been fetched
    def _booked_now_by_login(self, login):
        # must have run fetch_all() before calling this
        return any([lease.booked_now_by(self.leases_hostname, login)
                    for lease in self.leases])

    def _booked_now_by_anyone(self):
        # must have run fetch_all() before calling this
        return any([lease.booked_now_by(self.leases_hostname, login=None)
                    for lease in self.leases])

    async def fetch_all(self):
        """
        makes sure all required data is available 
        (turns out only the leases are required for now)
        """
        await self.fetch_leases()

    async def refresh(self):
        self.leases = None
        await self.fetch_all()

    def sort_leases(self):
        self.leases.sort(key=Lease.sort_key)

    async def fetch_leases(self):
        if self.leases is not None:
            return self.leases
        await self._fetch_leases()
        return self.leases

    # xxx stolen from r2lab.inria.fr : because we still use
    # a data model inspired from when we had OMF
    @staticmethod
    def epoch_to_ui_ts(epoch):
        return time.strftime("%Y-%m-%dT%H:%M:%SZ", time.gmtime(epoch))

    def resource_from_lease(self, plc_lease):
        return {'uuid': plc_lease['lease_id'],
                'slicename': plc_lease['name'],
                'valid_from': self.epoch_to_ui_ts(plc_lease['t_from']),
                'valid_until': self.epoch_to_ui_ts(plc_lease['t_until']),
                'ok': True}

    async def _fetch_leases(self):
        self.leases = None
        try:
            logger.info("Leases are being fetched..")
            self.plc_leases = self.plcapi_proxy.GetLeases(
                {'day': 0}, anonymous=True)
            logger.info("{} leases received".format(len(self.plc_leases)))
            # decoded as a list of Lease objects
            self.leases = [Lease(resource) for resource in self.plc_leases]
            self.sort_leases()
            self.fetch_time = time.strftime("%Y-%m-%d @ %H:%M")
            self.resources = [
                self.resource_from_lease(plc_lease)
                for plc_lease in self.plc_leases
            ]

        except Exception as e:
            if debug:
                print("Leases.fetch: exception {}".format(e))
            traceback.print_exc()
            await self.feedback('leases_error',
                                'cannot get leases from {} - exception {}'
                                .format(self, e))

    # this can be used with a fake message queue, it's synchroneous
    def print(self):
        print(5 * '-', self,
              "with special privileges"
              if self.has_special_privileges() else "")
        if self.leases is not None:
            def rights_char(lease):
                # B like broken
                if lease.broken:
                    return 'B'
                # S like super user
                if self.has_special_privileges():
                    return 'S'
                # * for yes you can use it
                if lease.booked_now_by(self.leases_hostname, self.login):
                    return '*'
                return ' '
            for i, lease in enumerate(self.leases):
                print("{:3d} {}".format(
                    i + 1, lease.__repr__(rights_char=rights_char(lease))))

    # material to create and modify leases
    @staticmethod
    def to_epoch(input):
        if not input:
            return time.time()
        if isinstance(input, (int, float)):
            return input
        patterns = [
            # fill in year
            "%Y-{}:00%Z",  # "%Y-{}:00",
            "%Y-%m-{}:00%Z",  # "%Y-%m-{}:00",
            "%Y-%m-%dT{}:00%Z",  # "%Y-%m-%dT{}:00",
            "%Y-%m-%dT{}:00:00%Z",  # "%Y-%m-%dT{}:%M:00",
        ]

        for pattern in patterns:
            fill = time.strftime(pattern).format(input)
            try:
                struct_time = time.strptime(fill, Lease.wire_timeformat)
                return int(time.mktime(struct_time))
            except:
                pass

    def check_user(self, user):
        return os.path.exists("/home/{}".format(user))

    async def _add_lease(self, owner, input_from, input_until):
        if owner != 'root':
            if not self.check_user(owner):
                print("user {} not found under /home - giving up"
                      .format(owner))
                logger.error("Unknown user {}".format(owner))
                return
        t_from = Leases.to_epoch(input_from)
        if not t_from:
            print("invalid time from: {}".format(input_from))
            return
        t_until = Leases.to_epoch(input_until)
        if not t_until:
            print("invalid time until: {}".format(input_until))
            return
        # just making sure
        try:
            hostname = self.leases_hostname
            retcod = self.plcapi_proxy.AddLeases(
                [hostname], owner, t_from, t_until)
            if 'new_ids' in retcod:
                # do we want to automatically
                # recompute the index ?
                print("OK")
                # force next reload
                self.leases = None
            elif 'errors' in retcod and retcod['errors']:
                for error in retcod['errors']:
                    print("error: {}".format(error))
        except Exception as e:
            print('Error', "Cannot add lease - e={}".format(e))
            traceback.print_exc()
            pass

    def get_lease_by_rank(self, lease_rank):
        try:
            irank = int(lease_rank)
            return self.leases[irank - 1]
        except:
            pass

    async def _update_lease(self, lease_rank, input_from, input_until):
        update_fields = {}
        if not input_from and not input_until:
            logger.info("update_lease : nothing to do")
            return
        if input_from:
            t_from = Leases.to_epoch(input_from)
            if not t_from:
                print("invalid time from: {}".format(input_from))
                return
            update_fields['t_from'] = t_from
        if input_until:
            t_until = Leases.to_epoch(input_until)
            if not t_until:
                print("invalid time until: {}".format(input_until))
                return
            update_fields['t_until'] = t_until
        # lease_rank could be a rank as displayed by self.print()
        the_lease = self.get_lease_by_rank(lease_rank)
        if not the_lease:
            print("Cannot find lease with rank {}".format(lease_rank))
            return
        lease_ids = [the_lease.lease_id]
        retcod = self.plcapi_proxy.UpdateLeases(lease_ids, update_fields)
        if 'errors' in retcod and retcod['errors']:
            for error in retcod['errors']:
                print("error: {}".format(error))
        else:
            print("OK")
            # force next reload
            self.leases = None

    async def _delete_lease(self, lease_rank):
        # lease_rank could be a rank as displayed by self.print()
        the_lease = self.get_lease_by_rank(lease_rank)
        if not the_lease:
            print("Cannot find lease with rank {}".format(lease_rank))
            return
        lease_ids = [the_lease.lease_id]
        retcod = self.plcapi_proxy.DeleteLeases(lease_ids)
        if retcod == 1:
            print("OK")
            # force next reload
            self.leases = None
        else:
            print("not deleted")

    async def main(self, interactive):
        await self.fetch_all()
        self.print()
        if not interactive:
            return 0
        try:
            result = await self.interactive()
            return result
        except (KeyboardInterrupt, EOFError) as e:
            print("Bye")
            return 1

    async def interactive(self):
        help_message = """
Enter one of the letters inside [], and answer the questions

A lease index is a number as shown on the left in the leases list

Times can be entered simply as
* just 14, or 14:00, for today at 2p.m.
* 14:30 for today at 2:30 p.m., or
* 27T10:30 for the 27th this month at 10:30 a.m., or
* 12-10T01:00 for the 12th december this year at 1 a.m., or
* 2016-01-02T08:00 for January 1st, 2016, at 8:00

In all the above cases, times will be understood as local time (France).

Leaving a time empty means either 'now', or 'do not change',
depending on the context
"""
        # interactive mode
        while True:
            current_time = time.strftime("%H:%M")
            answer = input("{} - Enter command "
                           "([l]ist, [a]dd, [u]pdate, "
                           "[d]elete, [r]efresh, [h]elp, [q]uit : "
                           .format(current_time))
            char = answer[0].lower() if answer else 'l'
            if char == 'l':
                self.print()
            elif char == 'a':
                if self.has_special_privileges():
                    owner = input("For slice name : ")
                else:
                    owner = self.login
                time_from = input("From : ")
                time_until = input("Until : ")
                print("owner={}".format(owner))
                result = await self._add_lease(owner, time_from, time_until)
                await self.fetch_all()
            elif char == 'u':
                rank = input("Enter lease index : ")
                time_from = input("From : ")
                time_until = input("Until : ")
                result = await self._update_lease(rank, time_from, time_until)
                await self.fetch_all()
            elif char == 'd':
                rank = input("Enter lease index : ")
                result = await self._delete_lease(rank)
                await self.fetch_all()
            elif char == 'r':
                await self.refresh()
                self.print()
            elif char == 'h':
                print(help_message)
            elif char == 'q':
                print('bye')
                break
            else:
                print("Command not understood {}".format(answer))
        return 0



#!/usr/bin/env python3
# determines local IP address to use for frisbeed

import subprocess
import re
import ipaddress

from rhubarbe.logger import logger

matcher = re.compile(r"inet (?P<address>([0-9]+\.){3}[0-9]+)/(?P<mask>[0-9]+)")

_local_interfaces = None

def local_interfaces():
    global _local_interfaces
    if _local_interfaces is not None:
        return _local_interfaces
    ip_links = subprocess.Popen(
        ['ip', 'address', 'show'],
        stdout = subprocess.PIPE,
        stderr = subprocess.DEVNULL,
        universal_newlines = True
    )
    _local_interfaces = []
    out, err = ip_links.communicate()
    for line in out.split("\n"):
        line = line.strip()
        m = matcher.match(line)
        if m:
            address, mask = m.group('address'), m.group('mask')
            interface = ipaddress.ip_interface("{}/{}".format(address, mask))
            if not interface.is_loopback:
                _local_interfaces.append(interface)
    return _local_interfaces

def local_ip_on_same_network_as(peer):
    """
    Typically if peer is 192.168.3.1 and we have an interface 192.168.3.200/24
    then this will return a tuple of strings 192.168.3.200, 24

    """
    #peer_address = ipaddress.ip_address(peer)
    for interface in local_interfaces():
        n = interface.network
        l = n.prefixlen
        peer_interface = ipaddress.ip_interface("{}/{}".format(peer,l))
        if peer_interface.network == interface.network:
            return str(interface.ip), str(l)
    
if __name__ == '__main__':
    local_ip, mask = local_ip_on_same_network_as("192.168.3.1")
    print("found {}/{}".format(local_ip, mask))



import os
import socket
import configparser
from pathlib import Path

from rhubarbe.singleton import Singleton
from rhubarbe.logger import logger


# location, mandatory

# all the files found in these locations are considered
# and are all loaded in this order, so the last ones
# overwrite the first ones
locations = [
    # path, mandatory
    ("/etc/rhubarbe/rhubarbe.conf", True),
    ("/etc/rhubarbe/rhubarbe.conf.local", False),
    (os.path.expanduser("~/.rhubarbe.conf"), False),
    ("./rhubarbe.conf", False),
    ("./rhubarbe.conf.local", False),
]


class ConfigException(Exception):
    pass


class Config(metaclass=Singleton):

    def __init__(self):
        self.parser = configparser.ConfigParser()
        self.files = []
        # load all configurations when they exist
        for location, mandatory in locations:
            if os.path.exists(location):
                self.files.append(location)
                self.parser.read(location)
                logger.info("Loaded config from {}".format(location))
            elif mandatory:
                raise ConfigException("Missing mandatory config file {}".format(location))
        #
        self._hostname = None

    def local_hostname(self):
        if not self._hostname:
            self._hostname = socket.gethostname().split('.')[0]
        return self._hostname

    def get_or_raise(self, dict, section, key):
        res = dict.get(key, None)
        if res is not None:
            return res
        else:
            raise ConfigException("rhubarbe config: missing entry section={} key={}"
                                  .format(section, key))

    def value(self, section, flag):
        if section not in self.parser:
            raise ConfigException("No such section {} in config".format(section))
        config_section = self.parser[section]
        hostname = self.local_hostname()
        return config_section.get("{flag}.{hostname}".format(**locals()), None) \
                or self.get_or_raise(config_section, section, flag)

    # for now
    # the foreseeable tricky part is, this should be a coroutine..
    def available_frisbee_port(self):
        return self.value('networking', 'port')

    # maybe this one too
    def local_control_ip(self):
        ### if specified in the config file, then use that
        if 'networking' in self.parser and 'local_control_ip' in self.parser['networking']:
            return self.parser['networking']['local_control_ip']
        ### but otherwise guess it
        # do not import at toplevel to avoid import loop
        from rhubarbe.inventory import Inventory
        the_inventory = Inventory()
        from rhubarbe.guessip import local_ip_on_same_network_as
        ip, prefixlen = local_ip_on_same_network_as(the_inventory.one_control_interface())
        return ip

    def display(self, sections):
        for i, file in enumerate(self.files):
            print("{}-th config file = {}".format(i+1, file))
        def match(section, sections):
            return not sections or section in sections
        for sname, section in sorted(self.parser.items()):
            if match(sname, sections) and section:
                print(10*'=', " section {}".format(sname))
                for fname, value in sorted(section.items()):
                    print("{} = {}".format(fname, value))


    def check_file_in_path(self, binary):
        PATH = os.environ['PATH']
        paths = ['/'] + [p for p in PATH.split(':') if p]
        for path in paths:
            full = Path(path) / binary
            if full.exists():
                return Path(path) / binary
        return False

    def check_binaries(self):
        # imagezip and frisbee are required on the pxe image only
        names = ('server', 'netcat')
        binaries = [self.value('frisbee', name) for name in names]

        for binary in binaries:
            checked = self.check_file_in_path(binary)
            if not checked:
                message = "Binary {} not found in PATH".format(binary)
                logger.critical(message)
                raise Exception(message)
            else:
                print("Found binary {} as {}"
                      .format(binary, checked))



import asyncio

from asynciojobs import Scheduler, Job

import rhubarbe.util as util
from rhubarbe.frisbeed import Frisbeed
from rhubarbe.leases import Leases
from rhubarbe.config import Config

class ImageLoader:

    def __init__(self, nodes,  image, bandwidth,
                 message_bus, display):
        self.nodes = nodes
        self.image = image
        self.bandwidth = bandwidth
        self.display = display
        self.message_bus = message_bus
        #
        self.frisbeed = None

    async def feedback(self, field, msg):
        await self.message_bus.put({field: msg})

    async def stage1(self):
        the_config = Config()
        idle = int(the_config.value('nodes', 'idle_after_reset'))
        await asyncio.gather(*[node.reboot_on_frisbee(idle) for node in self.nodes])

    async def start_frisbeed(self):
        self.frisbeed = Frisbeed(self.image, self.bandwidth, self.message_bus)
        ip_port = await self.frisbeed.start()
        return ip_port

    async def stage2(self, reset):
        """
        wait for all nodes to be telnet-friendly
        then run frisbee in all of them
        and reset the nodes afterwards, unless told otherwise
        """
        # start_frisbeed will return the ip+port to use 
        ip, port = await self.start_frisbeed()
        results = await asyncio.gather(*[node.run_frisbee(ip, port, reset) for node in self.nodes])
        # we can now kill the server
        self.frisbeed.stop_nowait()
        return all(results)

    # this is synchroneous
    def nextboot_cleanup(self):
        """
        Remove nextboot symlinks for all nodes in this selection
        so next boot will be off the harddrive
        """
        [node.manage_nextboot_symlink('harddrive') for node in self.nodes]

    async def run(self, reset):
        leases = Leases(self.message_bus)
        await self.feedback('authorization','checking for a valid lease')
        valid = await leases.booked_now_by_me()
        if not valid:
            await self.feedback('authorization',
                                "Access refused : you have no lease on the testbed at this time")
            return False
        else:
            await self.feedback('authorization','access granted')
            await (self.stage1() if reset else self.feedback('info', "Skipping stage1"))
            return await (self.stage2(reset))

    def main(self, reset, timeout):

        mainjob = Job(self.run(reset), critical=True)
        displayjob = Job(self.display.run(), forever=True, critical=True)
        scheduler = Scheduler (mainjob, displayjob)

        try:
            ok = scheduler.orchestrate(timeout = timeout)
            if not ok:
                scheduler.debrief()
                self.display.set_goodbye("rhubarbe-load failed: {}".format(scheduler.why()))
                return 1
            return 0 if mainjob.result() else 1
        except KeyboardInterrupt as e:
            self.display.set_goodbye("rhubarbe-load : keyboard interrupt - exiting")
            return 1
        finally:
            self.frisbeed and self.frisbeed.stop_nowait()
            self.nextboot_cleanup()
            self.display.epilogue()



import time
import random
import asyncio
import telnetlib3

from rhubarbe.logger import logger
from rhubarbe.config import Config


"""
TelnetProxy is what controls the telnet connection to the node,
it is subclassed as Frisbee and ImageZip

As per the design of telnetlib3, it uses a couple helper classes
* the shell (telnetlib3.TerminalShell) is what receives the session's output,
  so we have for example a FrisbeeParser class that acts as such a shell
* a client (telnetlib3.TelnetClient); 
  that we specialize so we can propagate our own stuff
  (the telnetproxy instance primarily) down to FrisbeeParser

This class essentially is subclassed as classes Frisbee and ImageZip  
"""

# one painful trick here is, we need to pass the shell class when connecting, even though
# in our usage model it would be more convenient to define this when the command is run

class TelnetClient(telnetlib3.TelnetClient):
    """
    this specialization of TelnetClient is meant for FrisbeeParser
    to retrieve its correponding TelnetProxy instance
    """
    def __init__(self, proxy, *args, **kwds):
        self.proxy = proxy
        super().__init__(*args, **kwds)


class TelnetProxy:
    """
    a convenience class that help us
    * wait for the telnet server to come up
    * invoke frisbee
    """
    def __init__(self, control_ip, message_bus):
        self.control_ip = control_ip
        self.message_bus = message_bus
        the_config = Config()
        self.port = int(the_config.value('networking', 'telnet_port'))
        self.backoff = float(the_config.value('networking', 'telnet_backoff'))
        self.timeout = float(the_config.value('networking', 'telnet_timeout'))
        # internals
        self._transport = None
        self._protocol = None

    def is_ready(self):
        # xxx for now we don't check that frisbee is installed and the expected version
        return self._protocol is not None

    async def feedback(self, field, msg):
        await self.message_bus.put({'ip': self.control_ip, field: msg})

    async def _try_to_connect(self, shell=telnetlib3.TerminalShell):
        
        # a little closure to capture our ip and expose it to the parser
        def client_factory():
            return TelnetClient(proxy = self, encoding='utf-8', shell=shell)

        await self.feedback('frisbee_status', "trying to telnet..")
        logger.info("Trying to telnet to {}".format(self.control_ip))
        loop = asyncio.get_event_loop()
        try:
            self._transport, self._protocol = \
              await asyncio.wait_for(
                  loop.create_connection(client_factory, self.control_ip, self.port),
                  self.timeout)
            logger.info("{}: telnet connected".format(self.control_ip))
            return True
        except asyncio.TimeoutError as e:
            await self.feedback('frisbee_status', "timed out..")
            self._transport, self._protocol = None, None            
        except Exception as e:
#            import traceback
#            traceback.print_exc()
            logger.info("telnet connect: unexpected exception {}".format(e))
            self._transport, self._protocol = None, None

    async def _wait_until_connect(self, shell=telnetlib3.TerminalShell):
        """
        wait for the telnet server to come up
        this has no native timeout mechanism
        """
        while True:
             await self._try_to_connect(shell)
             if self.is_ready():
                 return True
             else:
                 backoff = self.backoff*(0.5 + random.random())
                 await self.feedback(
                     'frisbee_status',
                     "backing off for {:.3}s".format(backoff))
                 await asyncio.sleep(backoff)



import json

from rhubarbe.singleton import Singleton
from rhubarbe.config import Config

class Inventory(metaclass = Singleton):

    def __init__(self):
        the_config = Config()
        with open(the_config.value('testbed', 'inventory_nodes_path')) as feed:
           self._nodes = json.load(feed)

    def _locate_entry_from_key(self, key, value):
        """
        search for an entry that has given hostname
        returns a tuple (host, key)
        e.g.
        _locate_entry_from_key('hostname', 'reboot01') =>
         ( { 'cmc' : {...}, 'control' : {...}, 'data' : {...} }, 'cmc' )
         """
        for host in self._nodes:
            for k, v in host.items():
                if v[key] == value:
                    return host, k
        return None, None

    def attached_hostname_info(self, hostname, interface_key='control', info_key='hostname'):
        """
        locate the entry that has at least one hostname equal to 'hostname'
        and returns the 'hostname' attached to that key
        e.g.
        attached_hostname('reboot01', 'control') => 'fit01'
        """
        host, k = self._locate_entry_from_key('hostname', hostname)
        if host and interface_key in host:
            return host[interface_key][info_key]

    def control_ip_from_any_ip(self, ip):
        host, k = self._locate_entry_from_key('ip', ip)
        if host:
            return host['control']['ip']

    def display(self, verbose=False):
        def cell_repr(k, v, verbose):
            if not verbose:
                return "{}:{}".format(k, v['hostname'])
            else:
                return "{}:{}[{}]".format(k, v['hostname'], v['mac'])
        print(20*'-', "INVENTORY CONTENTS")
        for node in self._nodes:
            print(" ".join([cell_repr(k, v, verbose) for k, v in node.items()]))
        print(20*'-', "INVENTORY END")

    def one_control_interface(self):
        return self._nodes[0]['control']['ip']

    def all_control_hostnames(self):
        return (node['control']['hostname'] for node in self._nodes)



import sys
import os
import os.path
import glob
import time
import re

from rhubarbe.config import Config
from rhubarbe.singleton import Singleton

debug = False

# use __ instead of == because = ruins bash completion
saving_keyword = 'saving'
saving_sep = '__'
saving_time_format = "%Y-%m-%d@%H-%M"

sep_re_pattern = "[-_=]{2}"
date_re_pattern = "[0-9]{4}-[0-9]{2}-[0-9]{2}@[0-9]{2}.[0-9]{2}"
# this will be fed into .format() - needs double {{ and }}
hostname_re_pattern_f = "{regularname}[0-9]{{2}}"

sep_glob_pattern = "[-_=]" * 2
date_glob_pattern = "[0-9-@]" * 16
hostname_glob_pattern_f = "{regularname}[0-9][0-9]"

class ImagesRepo(metaclass = Singleton):
    def __init__(self):
        the_config = Config()
        self.repo = the_config.value('frisbee', 'images_dir')
        self.default_name = the_config.value('frisbee', 'default_image')
        # for building hostname patterns
        self.regularname = the_config.value('testbed', 'regularname')

    suffix = ".ndz"

    def default(self):
        return os.path.join(self.repo, self.default_name)

    def add_extension(self, file):
        return file + self.suffix

    def where_to_save(self, nodename, radical):
        """
        given a nodename, plus an option user-provided image name (may be None)
        computes the actual path where to store an image
        * behaviour depends on actual id (root stores in global repo, regular users store in '.')
        * name always contains nodename and date
        """
        parts = [ saving_keyword, nodename, time.strftime(saving_time_format), radical ]
        base = saving_sep.join(parts) + self.suffix

        # root stores stuff directly in the repo
        if os.getuid() == 0:
            return os.path.join(self.repo, base)
        # regular users store where they are
        else:
            # a more sophisticated approach would be to split radical
            # to find about another directory, but well..
            return base

    def radical_part(self, input):
        """
        input can be a filename, possibly without an extension
        the saving__ blabla is extracted and only the part
        given at save-time is returned
        it is kind of the reverse of where_to_save
        """
        input = os.path.basename(input)
        input = os.path.splitext(input)[0]
        parts = [ saving_keyword,
                  hostname_glob_pattern_f.format(regularname = self.regularname),
                  date_glob_pattern,
                  "(?P<radical>.+)" ]
        glob_pattern = re.compile( saving_sep.join(parts))
        match = glob_pattern.match(input)
        if match:
            return match.group('radical')
        else:
            return input

    class ImageInfo:
        def __init__(self, repo, filename):
            # a valid path
            self.repo = repo
            self.filename = filename
            self.fill_info()

        def fill_info(self):
            self.radical = self.repo.radical_part(self.filename)
            stat = os.stat(self.filename)
            self.mtime = stat.st_mtime
            self.size = stat.st_size
            self.basename = os.path.splitext(os.path.basename (self.filename))
            self.is_official = self.radical == self.basename
            self.is_latest = None

        def __str__(self):
            return "{}{} {} {}".format(
                "*" if self.is_latest else " ",
                "O" if self.is_official else " ",
                self.filename,
                ImagesRepo.bytes2human(self.size))

    def _locate_infos(self, input_name, look_in_global):
        """
        search an image from its name
        * input_name:   a filename, possibly witout extension, possibly without saving_blabla
        * look_in_global: do we search in the global repo or just in .

        returns a list of ImageInfo objects
        (filename, radical, is_default, is_latest)
        where
        * filename is a valid filename that can be open()
        * radical 
        is_default is a boolean that says if this is the

        If the image name actually denotes a filename (i.e. with .ndz) then
        no completion takes place

        Otherwise, all 'saving' images are also considered and the latest one is picked
        """

        # when look_in_global is set, but we are already in the repo:
        # turn it off
        if look_in_global and os.path.abspath(os.getcwd()) == os.path.abspath(self.repo):
            look_in_global = False

        # existing filename ?
        candidate = input_name
        if os.path.exists(candidate):
            return [ ImagesRepo.ImageInfo(self, candidate) ]
        if look_in_global and not os.path.isabs(input_name):
            candidate = os.path.join(self.repo, input_name)
            if os.path.exists(candidate):
                return [ ImagesRepo.ImageInfo(self, candidate) ]

        # it's not a plain filename, so explore more options
        glob_pattern = saving_keyword + saving_sep + \
                       hostname_glob_pattern_f.format(regularname = self.regularname) + \
                       saving_sep + date_glob_pattern + saving_sep + input_name
        patterns = [ self.add_extension(input_name), self.add_extension(glob_pattern) ]
        if debug:
            print("glob_pattern", glob_pattern)
        if look_in_global and not os.path.isabs(input_name):
            repo_file = os.path.join(self.repo, input_name)
            repo_pattern = os.path.join(self.repo, glob_pattern)
            patterns += [ self.add_extension(repo_file), self.add_extension(repo_pattern) ]
        # run these patterns
        found = []
        for pattern in patterns:
            found += glob.glob(pattern)
        return [ ImagesRepo.ImageInfo(self, filename) for filename in found ]

    def _locate_infos_sorted(self, input_name, look_in_global=True):
        details = self._locate_infos(input_name, look_in_global)
        details.sort(key = lambda info: info.mtime, reverse=True)
        if details:
            details[0].is_latest = True
        return details

    def is_root(self):
        # surprisingly, even when running under sudo we have all 3 set to 0
        r, e, s = os.getresuid()
        is_root = r==0 and e==0 and s==0
        return is_root

    def locate_image(self, image_name, look_in_global=None):
        if look_in_global is None:
            look_in_global = self.is_root()
        infos = self._locate_infos_sorted(image_name, look_in_global)
        return infos[0].filename if infos else []

    @staticmethod
    def bytes2human(n, format="{value:.3f} {symbol}"):
        """
        >>> bytes2human(10000)
        '9K'
        >>> bytes2human(100001221)
        '95M'
        """
        symbols = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
        prefix = {}
        for i, s in enumerate(symbols[1:]):
            prefix[s] = 1 << (i+1)*10
        for symbol in reversed(symbols[1:]):
            if n >= prefix[symbol]:
                value = float(n) / prefix[symbol]
                return format.format(value=value, symbol=symbol)
        return format.format(symbol=symbols[0], value=n)

    # entry point for rhubarbe images
    def main(self, focus, verbose=False, sort_by='date', reverse=False):
        self.display(focus, verbose=verbose, sort_by=sort_by, reverse=reverse,
                     human_readable=True)

    # rhubarbe resolve
    def resolve(self, focus, verbose, reverse):
        for input in focus:
            infos =  self._locate_infos_sorted(input)
            if not infos:
                print("Image {} not found".format(input), file=sys.stderr)
            else:
                if not verbose:
                   # main purpose is to print out the right name
                    for info in infos:
                        if info.is_latest:
                            print(info.filename)
                else:
                    print(8*'*', "image {}".format(input))
                    # it comes in the wrong order, latest first
                    if not reverse:
                        infos.reverse()
                    for info in infos:
                        print(info)


    def display(self, focus, verbose, sort_by, reverse, human_readable):
        # show available images in some sensible way
        #
        # (*) focus: a list of re patterns
        # (.) if empty, everything is displayed
        # (.) otherwise only files that match one of these names
        #        are displayed (together with their symlinks)
        # (*) sort_by, reverse: how to sort
        # (*) human_readable: bool, show raw size or use units like MB or Gb 
        # 
        # We want to show all the names attached to a given image through symlinks
        # this is why all the filenames are grouped by clusters;
        # there is one cluster per real image file
        #
        # so what we do is
        # (1) scan all files and gather them by clusters that point at the same file
        # for this we use a (inode, mtime) tuple, same value = same file
        # (2) for each cluster, we distinguish between
        # 'single' clusters, i.e. an iamge file without any symlink
        # 'annotated' clusters have at least one symlink
        #
        # default behaviour is to show only annotated images
        # when verbose is turned on, all clusters are shown

        print("==================== Images repository {}".format(self.repo))
        if not verbose:
            print("==== use with --verbose to see images without a symlink")
        if focus:
            print("==== images matching any of", " ".join(focus))

        # gather one info per file, with
        # a key (inode, mtime, size) (same key = same file),
        # prefix (filename, no extension),
        # isalias
        infos = []
        for path in glob.glob("{}/*.ndz".format(self.repo)):
            prefix = os.path.basename(path).replace(".ndz", "")
            try:
                stat = os.stat(path)
            except Exception as e:
                print("WARNING - Cannot stat {} (dangling symlink ?) - ignored".format(path))
                continue
            stat_tuple = (stat.st_size, stat.st_mtime, stat.st_ino, )
            info = { 'stat_tuple' : stat_tuple, 'prefix' : prefix,
                     'isalias' : os.path.islink(path) }
            infos.append(info)
        # gather by same file (same stat_tuple)
        clusters = {}
        for info in infos:
            stat_tuple = info['stat_tuple']
            clusters.setdefault(stat_tuple, [])
            clusters[stat_tuple].append(info)

        # sort infos in one cluster, so that real files show up first
        sort_info = lambda info: info['isalias']
        for cluster in clusters.values():
            cluster.sort(key=sort_info)

        # sort clusters
        cluster_items = list(clusters.items())
        # k_v is a tuple with a key and a list of infos
        # k_v[0] is a key
        if sort_by == 'size':
            sort_clusters = lambda k_v: k_v[0][0]
        else:
            sort_clusters = lambda k_v: k_v[0][1]
        cluster_items.sort(key=sort_clusters, reverse=reverse)

        # is a given cluster filtered by focus
        def in_focus(cluster, focus):
            # an empty focus means to show all
            if not focus:
                return True
            # in this context a cluster is just a list of infos
            for info in cluster:
                for filter in focus:
                    # rough : no regexp, just find the filter or not
                    if info['prefix'].find(filter) >= 0:
#                        print("found match of {} in {}".
#                              format(filter, info['prefix']))
                        return True
                
        for stat_tuple, cluster in cluster_items:
            # skip the cluster if not in focus
            if not in_focus(cluster, focus):
                continue
            # without the verbose flag, show only clusters that have at least one symlink
            if not verbose:
                # do we have at least one symlink ?
                if not any([info['isalias'] for info in cluster]):
                    # ignoring single clusters in non-verbose mode 
                    continue
            shown = False
            for info in cluster:
                (size, mtime, inode) = info['stat_tuple']
                print_size = size if not human_readable \
                             else ImagesRepo.bytes2human(size)
                date = time.strftime("%Y-%m-%d %H:%M", time.localtime(mtime))
                prefix = info['prefix']
                if not shown:
                    print("{:<18}{:>12}B {:40}".format(date, print_size, prefix))
                    shown = True
                else:
                    print("{:>31} {:40}".format("a.k.a.", prefix))


    def share(self, images, alias, dry_run, force, clean):
        """
        A utility to install one or several images in the shared repo

        parameters:
        * images: a list of image names (not searched in global repo) to transfer
        * alias: a name for the installed image - only if a single image is provided
        
        Each provided image gets renamed into the radical of origin name, 
        i.e. with the extra saving__*blabla removed

        (.) When a single image is given, and 'alias' is set, then a symlink from alias to 
        radical is created

        (.) identity
        root is allowed to override destination, not regular users

        # return 0 if all goes well, 1 otherwise
        """

        ### check args
        if len(images) > 1 and alias:
            print("alias can be specified only with one image")
            return 1

        is_root = self.is_root()
        # compute dry_run if not set on the command line
        if dry_run is None:
            dry_run = False if is_root else True
        
        ### first pass
        # compute a list of tuples (origin, destination)
        for image in images:
            moves = []
            removes = [] # list of tuples 
            symlinks = [] # list of tuples
            origin = self.locate_image(image, look_in_global=is_root)
            if not origin:
                print("Could not find image {} - ignored".format(image))
                continue
            matching_infos = self._locate_infos_sorted(self.radical_part(origin),
                                                       look_in_global=is_root)
            # at this point there's at most one that is marked official
            # but if the user has given us a full path, that's the one
            # that is going to be official, so must not be removed either
            for info in matching_infos:
                if info.filename == origin:
                    info.is_official = True
            
            radical = self.radical_part(origin)
            if not radical:
                print("WARNING: Could not guess radical part in {}\n"
                      "  you will need to give one with -d".format(origin))
                continue
            destination = os.path.join(self.repo, radical + ".ndz")
            if os.path.exists(destination) and not force:
                print("WARNING: Destination {} already exists - ignored".format(destination))
            else:
                moves.append( (origin, destination) )
            if alias:
                symlinks.append( (radical + ".ndz", os.path.join(self.repo, alias + ".ndz")) )
            else:
                print("Warning : share without an alias")
                pass
            if clean:
                # item # 0 is the one selected for being moved
                for info in matching_infos[1:]:
                    # among the other ones, DON't remove the official since it's
                    # useless - or harmful
                    if not info.is_official:
                        removes.append(info.filename)

            print(8*'*', image)
            matching_infos.reverse()
            for info in matching_infos:
                print(info)

            def do_dry_run(*args):
                print("DRY-RUN: would do")
                print(*args)

            for remove in removes:
                if dry_run:
                    do_dry_run("rm {}".format(remove))
                else:
                    print("Removing {}".format(remove))
                    os.unlink(remove)
            
            for origin, destination in moves:
                if dry_run:
                    do_dry_run("mv {} {}".format(origin, destination))
                else:
                    print("Moving {} -> {}".format(origin, destination))
                    os.rename(origin, destination)

            for origin, destination in symlinks:
                if dry_run:
                    do_dry_run("ln -sf {} {}".format(origin, destination))
                else:
                    print("Creating symlink {} -> {}".format(destination, origin))
                    if os.path.exists(destination):
                        os.unlink(destination)
                    os.symlink(origin, destination)
        
        return 0



# this class allows to create an authenticated xmlrpc
# connection to 

import getpass

import ssl

#from aioxmlrpc.client import ServerProxy
from xmlrpc.client import ServerProxy

class PlcApiProxy(ServerProxy):

    # use the standard plcapi scheme to run on /PLCAPI/
    # always use password auth in this first rough version
    # could be improved by using session authentication
    def __init__(self, url, email=None, password=None, debug=False):
        self.url = url
        self.email = email
        self.password = password
        self.debug = debug
        ###
        context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
        context.check_hostname = False
        ServerProxy.__init__(
            self, self.url, allow_none=True, context=context
        )
        
    def __auth__(self, anonymous):
        if anonymous:
            return { 'AuthMethod' : 'anonymous' }
        else:
            if not self.email:
                self.email = input("Enter plcapi email (login) : ")
            if not self.password:
                self.password = getpass.getpass("Enter plcapi password for {} : "
                                                .format(self.email))
            return { 'AuthMethod' : 'password',
                     'Username'   : self.email,
                     'AuthString' : self.password }
        

    # pass the authentication along for all calls
    def __getattr__(self, attr):
        # the default is to use authenticated calls
        # because this is tha majority of the plcapi calls
        def fun(*args, anonymous=False, **kwds):
            if self.debug:
                auth_msg = "[auth]" if not anonymous else "[anon]"
                print("-> Sending {} {} on {} with args={} and kwds={}"
                      .format(auth_msg, attr, self, args, kwds))
            actual_fun = ServerProxy.__getattr__(
                self, attr)
            try:
                retcod = actual_fun(self.__auth__(anonymous), *args, **kwds)
                if self.debug:
                    print("<- Received {}".format(retcod))
                return retcod
            except Exception as e:
                print("ignored exception in {} : {}"
                          .format(attr, e))
        return fun

    def __str__(self):
        return "PLCAPIproxy@{}".format(self.url)



"""
This implements a daemon that runs on faraday and that

* cyclically fetches something like GetSlivers()
  and adjusts local accounts accordingly
* it is also planned that it will listen on some port
  so that one can get it to fuse its cycle an redo the
  core of its job right away

think of it as a dedicated nodemanager
"""

import time
import os.path
import pwd
import glob

from rhubarbe.logger import accounts_logger as logger
from .config import Config
from .plcapiproxy import PlcApiProxy

####################
# stolen from historical planetlab nodemanager


def replace_file_with_string(filename, new_contents,
                             owner=None, chmod=None, remove_if_empty=False):
    """
    Replace a file with new contents
    checks for changes
      does not do anything if previous state was already right
    can handle chmod/chown if requested
    can also remove resulting file if contents are void, if requested

    returns True if a change occurred, or the file is deleted
    """
    try:
        with open(filename) as f:
            current = f.read()
    except IOError:
        current = ""
    if current == new_contents:
        # if turns out to be an empty string, and remove_if_empty is set,
        # then make sure to trash the file if it exists
        if remove_if_empty and not new_contents and os.path.isfile(filename):
            logger.info(
                "replace_file_with_string: removing file {}".format(filename))
            try:
                os.unlink(filename)
            finally:
                return True
        # we're done and have nothing to do
        return False
    # overwrite filename file: create a temp in the same directory
    with open(filename, 'w') as f:
        f.write(new_contents)
    if chmod:
        os.chmod(filename, chmod)
    if owner:
        os.system("chown {} {}".format(owner, filename))
    return True

####################


class Accounts:

    def __init__(self):
        the_config = Config()
        self.plcapiurl = the_config.value('plcapi', 'url')
        self.email = the_config.value('plcapi', 'admin_email')
        self.password = the_config.value('plcapi', 'admin_password')

        self._proxy = None

    # not reconnectable for now
    def proxy(self):
        if self._proxy is None:
            self._proxy = PlcApiProxy(self.plcapiurl,
                                      email=self.email, password=self.password,
                                      # debug = True
                                     )
        return self._proxy

    def authorized_home_basenames(self):
        """
        Inspect /home/ to find accounts that have authorized_keys

        temporarily leave alone the ones that start with 'onelab.'

        also focus on the ones that have a '_' in them so that we leave
        artificial accounts like 'michelle' or 'mario' that can maybe
        be helpful some day

        """
        basenames = []
        for homedir in glob.glob("/home/*"):
            basename = os.path.basename(homedir)
            if basename.startswith("onelab."):
                continue
            if '_' not in basename:
                continue
            if not os.path.exists(
                    os.path.join(homedir, ".ssh/authorized_keys")):
                continue
            basenames.append(basename)
        return basenames

    def authorized_legacy_basenames(self):
        """
        Inspect /home/ to find accounts that have authorized_keys
        focus on the ones that start with 'onelab.'

        """
        basenames = []
        for homedir in glob.glob("/home/onelab.*"):
            basename = os.path.basename(homedir)
            if not os.path.exists(
                    os.path.join(homedir, ".ssh/authorized_keys")):
                continue
            basenames.append(basename)
        return basenames

    def all_passwd_entries(self):
        """
        Returns all entries in /etc/passwd
        this is just to figure if an account needs to be created
        """
        return [pw.pw_name for pw in pwd.getpwall()]

    def create_account(self, slicename):
        """
        Does useradd with the right options
        Plus, creates an empty .ssh dir with proper permissions
        NOTE that this addresses ubuntu for now, fedora 'useradd'
        being slightly different as far as I remember
        (at least wrt homedir creation, IIRC again)
        """
        commands = [
            "useradd --create-home --user-group {x} --shell /bin/bash",
            "mkdir /home/{x}/.ssh",
            "chmod 700 /home/{x}/.ssh",
            "chown -R {x}:{x} /home/{x}",
        ]
        for cmd in commands:
            command = cmd.format(x=slicename)
            logger.info("Running {}".format(command))
            retcod = os.system(command)
            if retcod != 0:
                logger.error("{} -> {}".format(command, retcod))

    def create_ssh_config(self, slicename):
        """
        Initialize slice's .ssh/config that keeps ssh from
        being too picky with host keys and similar

        Performed only if not yet existing
        """
        ssh_config_file = "/home/{x}/.ssh/config".format(x=slicename)

        if not os.path.exists(ssh_config_file):

            # define the magic sequence for both fit* and data*
            config_bases = ['fit', 'data']
            #
            config_pattern = """Host {base}*
    StrictHostKeyChecking no
    UserKnownHostsFile=/dev/null
    CheckHostIP=no
"""

            ssh_config = "\n".join(
                [config_pattern.format(base=base) for base in config_bases])
            replace_file_with_string(ssh_config_file,
                                     ssh_config,
                                     chmod=0o600,
                                     owner="{x}:{x}".format(x=slicename))

    ##########
    def authorized_key_lines(self, plc_slice,
                             plc_persons_by_id, plc_keys_by_id):
        """
        returns the expected contents of that slice's authorized_keys file
        """
        slicename = plc_slice['name']
        persons = [plc_persons_by_id[id] for id in plc_slice['person_ids']]
        key_ids = set(sum((p['key_ids'] for p in persons), []))
        keys = [plc_keys_by_id[id] for id in key_ids]
        key_lines = [(k['key'] + "\n") for k in keys]
        # this is so we get something canonical that does not
        # change everytime
        key_lines.sort()
        return "".join(key_lines)

    ##########
    def manage_accounts(self):
        # get context
        passwd_entries = self.all_passwd_entries()
        home_basenames = self.authorized_home_basenames()
        legacy_basenames = self.authorized_legacy_basenames()

        # get plcapi specification of what should be
        slices = self.proxy().GetSlices(
            {}, ['slice_id', 'name', 'expires', 'person_ids'])
        persons = self.proxy().GetPersons(
            {}, ['person_id', 'email', 'slice_ids', 'key_ids'])
        keys = self.proxy().GetKeys()

        if slices is None or persons is None or keys is None:
            logger.error("Cannot reach PLCAPI endpoint at this time - back to sleep")
            return

        persons_by_id = {p['person_id']: p for p in persons}
        keys_by_id = {k['key_id']: k for k in keys}

        active_slices = []

        for slice in slices:
            slicename = slice['name']
            authorized_keys = self.authorized_key_lines(
                slice, persons_by_id, keys_by_id)

            # don't bother to create an account if the slice has no key
            if authorized_keys:
                try:
                    active_slices.append(slicename)
                    # create account if missing
                    if slicename not in passwd_entries:
                        self.create_account(slicename)
                    # dictate authorized_keys contents
                    ssh_auth_keys = "/home/{x}/.ssh/authorized_keys".format(
                        x=slicename)
                    replace_file_with_string(ssh_auth_keys, authorized_keys,
                                             chmod=0o400,
                                             owner="{x}:{x}".format(
                                                 x=slicename),
                                             remove_if_empty=True)
                    self.create_ssh_config(slicename)

                except Exception as e:
                    logger.exception("could not properly deal "
                                     "with active slice {x} (e={e})"
                                     .format(x=slicename, e=e))

        # find out about slices that currently have suthorized keys but should
        # not
        for slicename in home_basenames:
            if slicename not in active_slices:
                try:
                    logger.info(
                        "Removing authorized_keys for {x}".format(x=slicename))
                    ssh_auth_keys = "/home/{x}/.ssh/authorized_keys".format(
                        x=slicename)
                    os.unlink(ssh_auth_keys)
                except Exception as e:
                    logger.exception("could not properly deal "
                                     "with inactive slice {x} (e={e})"
                                     .format(x=slicename, e=e))

        # a one-shot piece of code : turn off legacy slices
        for slicename in legacy_basenames:
            try:
                logger.info(
                    "legacy slicename {x} "
                    "needs to be shutdown".format(x=slicename))
                ssh_auth_keys = "/home/{x}/.ssh/authorized_keys".format(
                    x=slicename)
                # xxx enable the following line to tear down legacy slices
                # os.unlink(ssh_auth_keys)
            except Exception as e:
                logger.exception("could not properly deal "
                                 "with inactive slice {x} (e={e})"
                                 .format(x=slicename, e=e))


    def run_forever(self, period):
        while True:
            beg = time.time()
            logger.info("---------- rhubarbe accounts manager (period {})"
                        .format(period))
            self.manage_accounts()
            now = time.time()
            duration = now - beg
            towait = period - duration
            logger.info("---------- rhubarbe accounts manager - sleeping for {}"
                        .format(towait))
            if towait <= 0:
                logger.warning("duration {} exceeded period {} - skipping sleep"
                               .format(duration, period))
            else:
                time.sleep(period - duration)

    def main(self, cycle):
        if cycle is None:
            cycle = Config().value('accounts', 'cycle')
        cycle = int(cycle)
        if cycle != 0:
            self.run_forever(cycle)
        else:
            self.manage_accounts()
              



import os
import time
import re
import asyncio
import telnetlib3

from rhubarbe.logger import logger
from rhubarbe.config import Config
from rhubarbe.telnet import TelnetProxy

class ImageZipParser(telnetlib3.TerminalShell):
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.bytes_line = b""
        self.total_chunks = 0

    def feed_byte(self, x):
        if x == b"\n":
            self.parse_line()
            self.bytes_line = b""
        else:
            self.bytes_line += x

    def ip(self):
        return self.client.proxy.control_ip
    def feedback(self, field, msg):
        self.client.proxy.message_bus.put_nowait({'ip': self.ip(), field: msg})
    def send_percent(self, percent):
        self.feedback('progress', percent)

    # parse imagezip output ????
    def parse_line(self):
        line = self.bytes_line.decode().strip()
        logger.debug("line from imagezip:" + line)
        #
        # we don't parse anything here because there does not seem to be a way
        # to estimate some total first off, and later get percentages
        # useful to see the logs:
        # self.feedback('imagezip_raw', line)
        # send 10 ticks in a raw - not a good idea
        # for i in range(10): self.feedback('tick', '')
        
class ImageZip(TelnetProxy):
    async def connect(self):
        await self._try_to_connect(shell=ImageZipParser)

    async def wait(self):
        await self._wait_until_connect(shell=ImageZipParser)

    async def ticker(self):
        while self._running:
            await self.feedback('tick', '')
            await asyncio.sleep(0.1)

    async def wait_protocol_and_stop_ticker(self):
        await self._protocol.waiter_closed
        # hack so we can finish the progressbar
        await self.feedback('tick', 'END')
        self._running = False

    async def run(self, port, nodename, radical, comment):
        the_config = Config()
        server_ip = the_config.local_control_ip()
        imagezip = the_config.value('frisbee', 'imagezip')
        netcat = the_config.value('frisbee', 'netcat')
        # typically /dev/sda
        hdd = the_config.value('frisbee', 'hard_drive')
        # typically /dev/sda1
        root_partition = the_config.value('frisbee', 'root_partition')
        command = ""
        if root_partition and root_partition.lower() != 'none':
            # Managing the /etc/rhubarbe-image stamp
            # typically /mnt
            mount_point = the_config.value('frisbee', 'mount_point')
            date = time.strftime("%Y-%m-%d@%H:%M", time.localtime())
            who = os.getlogin()
            # create mount point if needed
            format = '[ -d {mount_point} ] || mkdir {mount_point}; '
            # mount it, and only if successful ...
            format += 'mount {root_partition} {mount_point} && '
            # add to the stamp, and umount - beware of {{ and }} as these are formats
            format += '{{ echo "{date} - node {nodename} - image {radical} - by {who}"'
            if comment:
                format += '" - {}"'.format(comment)
            format += ' >> {mount_point}/etc/rhubarbe-image ; umount {mount_point}; }} ; '
            # replace {} 
            command += format.format(**locals())
        command += \
          "{imagezip} -o -z1 {hdd} - | {netcat} {server_ip} {port}".format(**locals())

        logger.info("on {} : running command {}".format(self.control_ip, command))
        await self.feedback('frisbee_status', "starting imagezip on {}".format(self.control_ip))
        
        EOF = chr(4)
        EOL = '\n'
        # print out exit status so the parser can catch it and expose it
        command = command + "; echo FRISBEE-STATUS=$?"
        # make sure the command is sent (EOL) and that the session terminates afterwards (exit + EOF)
        command = command + "; exit" + EOL + EOF
        self._protocol.stream.write(self._protocol.shell.encode(command))

        # wait for telnet to terminate
        self._running = True
        await asyncio.gather(self.ticker(), self.wait_protocol_and_stop_ticker())
        return True



import time
import re
import json
import asyncio
from urllib.parse import urlparse

# turn off warnings that show up in monitor's journal
# https://urllib3.readthedocs.io/en/latest/advanced-usage.html#ssl-warnings
# worth a try someday..
import urllib3
urllib3.disable_warnings()

# to connect to sidecar
# at first I would have preferred an asyncio-friendly library
# for talking socket.io; however I could not find one, and
# websockets did not offer something like 'emit' out of the box
# so, we're still using this synchroneous one
# using .on to arm callbacks, and occasionally calling wait
# to give a chance for the callback to trigger
from socketIO_client import SocketIO, LoggingNamespace

from rhubarbe.config import Config

########## use a dedicated logger for monitor
from rhubarbe.logger import monitor_logger as logger
from rhubarbe.node import Node
from rhubarbe.ssh import SshProxy

###
# the channel that allows to fuse the leases acquisition cycle
# not too clean of course...
back_channel = Config().value('sidecar', 'channel_leases_request')

##########
class ReconnectableSocketIO:
    """
    can emit a message to a socketio service, or reconnect to it
    NOTE that this implementation is not robust, in the sense 
    that when we attempt to emit a message to a disconnected service,
    this message is dropped
    """

    def __init__(self, url, verbose=False):
        self.url = url
        # parse url
        parsed = urlparse(url)
        self.scheme, self.hostname, self.port \
            = parsed.scheme, parsed.hostname, parsed.port or 80
        self.verbose = verbose
        # internal stuff
        self.socketio = None
        self.counters = { 'connect' : 0 }

    def __repr__(self):
        return "socket.io server at {} ({} connections)"\
            .format(self.url, self.counters['connect'])
    
    # at some point we were running into the same issue as this one:
    # https://github.com/liris/websocket-client/issues/222
    # hopefully this harmless glitch has gone away
    def connect(self):
        action = "connect" if self.socketio is None else "reconnect"
        try:
            logger.info("{}ing to {}".format(action, self))
            # this might be due to miscconfiguration
            if self.scheme not in ('http', 'https'):
                logger.warn("unsupported scheme {} - malformed socketio URL: {}"
                            .format(self.scheme, self.url))
            if self.scheme == 'http':
                host_part = self.hostname
                extras = {}
            else:
                host_part = "https://{}".format(self.hostname)
                extras = {'verify' : False}
            self.socketio = SocketIO(host_part,
                                     self.port,
                                     LoggingNamespace,
                                     **extras)
            channel = back_channel
            def closure(*args):
                return self.on_channel(channel, *args)
            self.socketio.on(channel, closure)
            self.counters['connect'] += 1
            logger.info("{}ed to {}".format(action, self))
        except Exception as e:
            logger.warn("Connection lost to {} (e={})".format(self, e))
            logger.exception("Exception stack:")
            self.socketio = None

    def on_channel(self, channel, *args):
        # not supposed to run - should be redefined
        logger.warning("ReconnectableSocketIO.on_channel, channel={}, args={}"
                       .format(channel, args))

    def get_counter(self, channel):
        return self.counters.get(channel, 0)

    def emit_info(self, channel, info, wrap_in_list):
        # info can be a list (e.g. for leases)
        # or a dict, with or without an 'id' field
        if self.verbose:
            msg = "len={}".format(len(info)) if isinstance(info, list) \
                  else "id={}".format(info.get('id', '[none]'))
            logger.info("{} emitting {} -> {}"
                        .format(channel, msg, info))
                            
        # wrap info as single elt in a list
        emitted_info = info if not wrap_in_list else [ info ]
        message = json.dumps(emitted_info)
        if self.socketio is None:
            self.connect()
        try:
            self.socketio.emit(channel, message,
                               ReconnectableSocketIO.callback)
            self.counters.setdefault(channel, 0)
            self.counters[channel] += 1
        except Exception as e:
            # make sure we reconnect later on
            self.socketio = None
            label = "{}: {}".format(info.get('id', 'none'), one_char_summary(info))
            logger.warn("Dropped message on {} - channel {} - msg {}"
                        .format(self, channel, label))
            logger.exception("Dropped because of this exception:")
            
    def wait(self, wait):
        if self.socketio:
            try:
                self.socketio.wait(wait)
            except:
                pass

    @staticmethod
    def callback(*args, **kwds):
        logger.info('on socketIO response args={} kwds={}'.format(args, kwds))

class ReconnectableSocketIOMonitor(ReconnectableSocketIO):
    """
    A ReconnectableSocketIO 
    with a back link to a Monitor object, 
    that can receive the 'on_channel' method
    """

    def __init__(self, monitor, *args, **kwds):
        self.monitor = monitor
        ReconnectableSocketIO.__init__(self, *args, **kwds)

    def on_channel(self, channel, *args):
        """
        triggers on_channel back on monitor object
        """
        self.monitor.on_channel(channel, *args)

# translate info into a single char for logging
def one_char_summary(info):
    if 'cmc_on_off' in info and info['cmc_on_off'] != 'on':
        return '.'
    elif 'control_ping' in info and info['control_ping'] != 'on':
        return 'o'
    elif 'control_ssh' in info and info['control_ssh'] != 'on':
        return '0'
    elif 'os_release' in info and 'fedora' in info['os_release']:
        return 'F'
    elif 'os_release' in info and 'ubuntu' in info['os_release']:
        return 'U'
    else:
        return '^'

class MonitorNode:
    """
    the logic for probing a node as part of monitoring
    formerly in fitsophia/website/monitor.py
    . first probes for the CMC status; if off: done
    . second probe for ssh; if on : done
    . third, checks for ping
    """
    
    def __init__(self, node, reconnectable, channel, report_wlan=True, verbose=False):
        # a rhubarbe.node.Node instance
        self.node = node
        self.report_wlan = report_wlan
        self.reconnectable = reconnectable
        self.channel = channel
        self.verbose = verbose
        # current info - will be reported to sidecar
        self.info = {'id': node.id }
        # remember previous wlan measurement to compute rate
        self.history = {}

    def set_info(self, *overrides):
        """
        update self.info with all the dicts in overrides
        """
        for override in overrides:
            self.info.update(override)

    def zero_wlan_infos(self):
        """
        set wlan-related attributes to 0.
        """
        for k in self.info:
            if k.startswith('wlan'):
                self.info[k] = 0.
        
    def report_info(self):
        """
        Send info to sidecar
        """
        self.reconnectable.emit_info(self.channel, self.info, wrap_in_list=True)
        
    def set_info_and_report(self, *overrides):
        """
        Convenience to set info and immediately report
        """
        self.set_info(*overrides)
        self.report_info()

    ubuntu_matcher = re.compile("DISTRIB_RELEASE=(?P<ubuntu_version>[0-9.]+)")
    fedora_matcher = re.compile("Fedora release (?P<fedora_version>\d+)")
    gnuradio_matcher = re.compile("\AGNURADIO:(?P<gnuradio_version>[0-9\.]+)\Z")
    uname_matcher = re.compile("\AUNAME:(?P<uname>.+)\Z")
    # 2016-05-28@08:20 - node fit38 - image oai-enb-base2 - by root
    rhubarbe_image_matcher = re.compile("\A/etc/rhubarbe-image:" + \
                                        ".* - image (?P<image_radical>[^ ]+) - by"
                                        )
    rxtx_matcher = re.compile("==> /sys/class/net/wlan(?P<wlan_no>[0-9])/statistics/(?P<rxtx>[rt]x)_bytes <==")
    number_matcher = re.compile("\A[0-9]+\Z")

    def parse_ssh_probe_output(self, stdout, padding_dict):
        os_release = "other"
        gnuradio_release = "none"
        uname = ""
        rxtx_dict = {}
        rxtx_key = None
        image_radical = ""
        for line in stdout.split("\n"):
            match = self.ubuntu_matcher.match(line)
            if match:
                version = match.group('ubuntu_version')
                os_release = "ubuntu-{version}".format(**locals())
                continue
            match = self.fedora_matcher.match(line)
            if match:
                version = match.group('fedora_version')
                os_release = "fedora-{version}".format(**locals())
                continue
            match = self.gnuradio_matcher.match(line)
            if match:
                gnuradio_release = match.group('gnuradio_version')
                continue
            match = self.uname_matcher.match(line)
            if match:
                uname = match.group('uname')
                continue
            match = self.rhubarbe_image_matcher.match(line)
            if match:
                image_radical = match.group('image_radical')
                continue
            match = self.rxtx_matcher.match(line)
            if match:
                # use a tuple as the hash
                rxtx_key = (match.group('wlan_no'), match.group('rxtx'))
                continue
            match = self.number_matcher.match(line)
            if match and rxtx_key:
                rxtx_dict[rxtx_key] = int(line)
                continue
            rxtx_key = None
            
        # now that we have the counters we need to translate this into rate
        # for that purpose we use local clock; small imprecision should not impact overall result
        now = time.time()
        wlan_info_dict = {}
        for rxtx_key, bytes in rxtx_dict.items():
            wlan_no, rxtx = rxtx_key
            # rather dirty hack for images that use wlan2 and wlan3
            # expose in wlan0 or wlan1 depending on parity of actual device
            try:
                wlan_no = int(wlan_no)
                wlan_no = wlan_no % 2
            except:
                pass
            if self.verbose:
                logger.info("node={self.node} collected {bytes} for device wlan{wlan_no} in {rxtx}"
                            .format(**locals()))
            # do we have something on this measurement ?
            if rxtx_key in self.history:
                previous_bytes, previous_time = self.history[rxtx_key]
                info_key = "wlan_{wlan_no}_{rxtx}_rate".format(**locals())
                new_rate = 8.*(bytes - previous_bytes) / (now - previous_time)
                wlan_info_dict[info_key] = new_rate
                if self.verbose:
                    logger.info("node={} computed {} bps for key {} "
                                "- bytes = {}, pr = {}, now = {}, pr = {}"
                                .format(id, new_rate, info_key,
                                        bytes, previous_bytes, now, previous_time))
            # store this measurement for next run
            self.history[rxtx_key] = (bytes, now)
        # xxx would make sense to clean up history for measurements that
        # we were not able to collect at this cycle
        self.set_info( { 'os_release' : os_release,
                         'gnuradio_release' : gnuradio_release,
                         'uname' : uname,
                         'image_radical' : image_radical,
                     },
                       padding_dict, wlan_info_dict)

    async def probe(self, ping_timeout, ssh_timeout):
        """
        The logic for getting one node's info and send it to sidecar
        """
        if self.verbose:
            logger.info("entering pass1, info={}".format(self.info))
        # pass1 : check for status
        padding_dict = {
            'control_ping' : 'off',
            'control_ssh' : 'off',
            # don't overwrite os_release though
        }
        # get USRP status no matter what - use "" if we receive None
        # to limit noise when the node is physically removed
        usrp_status = await self.node.get_usrpstatus() or 'fail'
        # replace usrpon and usrpoff with just on and off
        self.set_info({'usrp_on_off' : usrp_status.replace('usrp', '')})
        # get CMC status
        status = await self.node.get_status()
        if status == "off":
            self.set_info_and_report({'cmc_on_off' : 'off'}, padding_dict)
            return
        elif status != "on":
            self.set_info_and_report({'cmc_on_off' : 'fail'}, padding_dict)
            return
        if self.verbose:
            logger.info("entering pass2, info={}".format(self.info))
        # pass2 : CMC status is ON - let's try to ssh it
        self.set_info({'cmc_on_off' : 'on'})
        padding_dict = {
            'control_ping' : 'on',
            'control_ssh' : 'on',
        }
        self.zero_wlan_infos()
        remote_commands = [
            "cat /etc/lsb-release /etc/fedora-release /etc/gnuradio-release 2> /dev/null | grep -i release",
            "echo -n GNURADIO: ; gnuradio-config-info --version 2> /dev/null || echo none",
            # this trick allows to have the filename on each output line
            "grep . /etc/rhubarbe-image /dev/null",
            "echo -n 'UNAME:' ; uname -r",
            ]
        if self.report_wlan:
            remote_commands.append(
                "head /sys/class/net/wlan?/statistics/[rt]x_bytes"                
            )
        # reconnect each time
        async with SshProxy(self.node) as ssh:
            if self.verbose:
                logger.info("trying to ssh-connect (timeout={})"
                            .format(ssh_timeout))
            try:
                connected = await asyncio.wait_for(ssh.connect(), timeout=ssh_timeout)
            except asyncio.TimeoutError as e:
                connected = False
            if self.verbose:
                logger.info("ssh-connected={}".format(connected))
            if connected:
                try:
                    command = ";".join(remote_commands)
                    output = await ssh.run(command)
                    # padding dict here sets control_ssh and control_ping to on
                    self.parse_ssh_probe_output(output, padding_dict)
                    # required as otherwise we leak openfiles
                    try:
                        await ssh.close()
                    except Exception as e:
                        logger.exception("monitor oops 1")
                        pass
                except Exception as e:
                    logger.exception("monitor remote_command failed")
            else:
                self.set_info({'control_ssh': 'off'})

        # if we could ssh then we're done
        if self.info['control_ssh'] == 'on':
            self.report_info()
            return

        if self.verbose:
            logger.info("entering pass3, info={}".format(self.info))
        # pass3 : node is ON but could not ssh
        # check for ping
        # I don't know of an asyncio library to deal with icmp
        # so let's use asyncio.subprocess
        # xxx maybe a Ping class would be the way to go
        control = self.node.control_hostname()
        command = [ "ping", "-c", "1", "-t", "1", control ]
        try:
            subprocess = await asyncio.create_subprocess_exec(
                *command,
                stdout = asyncio.subprocess.DEVNULL,
                stderr = asyncio.subprocess.DEVNULL)
            retcod = await asyncio.wait_for(subprocess.wait(), timeout=ping_timeout)
            self.set_info_and_report({'control_ping' : 'on'})
            return
        except asyncio.TimeoutError as e:
            self.set_info_and_report({'control_ping' : 'off'})
            return

    async def probe_forever(self, cycle, ping_timeout, ssh_timeout):
        """
        runs forever, wait <cycle> seconds between 2 runs of probe()
        """
        while True:
            try:
                await self.probe(ping_timeout, ssh_timeout)
            except Exception as e:
                logger.exception("monitor oops 2")
            await asyncio.sleep(cycle)
            

from rhubarbe.leases import Lease, Leases

class MonitorLeases:
    def __init__(self, message_bus, reconnectable, channel, cycle, step, wait, verbose):
        self.message_bus = message_bus
        self.reconnectable = reconnectable
        self.channel = channel
        self.cycle = float(cycle)
        self.step = float(step)
        self.wait = float(wait)
        self.verbose = verbose

    def on_back_channel(self, *args):
        # when anything is received on the backchannel, we just go to fast track
        logger.info("MonitorLeases.on_back_channel, args={}".format(args))
        self.fast_track = True

    async def run_forever(self):
        leases = Leases(self.message_bus)
        while True:
            self.fast_track = False
            trigger = time.time() + self.cycle
            # check for back_channel every 15 ms
            while not self.fast_track and time.time() < trigger:
                await asyncio.sleep(self.step)
                # give a chance to socketio events to trigger
                self.reconnectable.wait(self.wait)
                
            try:
                await leases.refresh()
                # xxx this is fragile
                omf_leases = leases.resources
                self.reconnectable.emit_info(self.channel, omf_leases,
                                             wrap_in_list=False)
                logger.info("advertising {} leases".format(len(omf_leases)))
                if self.verbose:
                    logger.info("Leases details: {}".format(omf_leases))
            except Exception as e:
                logger.exception("monitor could not get leases")
            
class Monitor:
    def __init__(self, cmc_names, message_bus, cycle, sidecar_url,
                 report_wlan=True, verbose=False):
        self.cycle = cycle
        self.report_wlan = report_wlan
        self.verbose = verbose

        # get miscell config
        self.ping_timeout = float(Config().value('networking', 'ping_timeout'))
        self.ssh_timeout = float(Config().value('networking', 'ssh_timeout'))
        self.log_period = float(Config().value('monitor', 'log_period'))

        # socket IO pipe
        self.reconnectable = ReconnectableSocketIOMonitor(self, sidecar_url, verbose)

        # the nodes part
        self.main_channel = Config().value('sidecar', 'channel_nodes')
        nodes = [ Node (cmc_name, message_bus) for cmc_name in cmc_names ]
        self.monitor_nodes = [
            MonitorNode(node = node, reconnectable = self.reconnectable,
                        channel = self.main_channel, report_wlan = self.report_wlan,
                        verbose=verbose)
            for node in nodes]

        # the leases part
        cycle = Config().value('monitor', 'cycle_leases')
        step = Config().value('monitor', 'step_leases')
        wait = Config().value('monitor', 'wait_leases')
        channel = Config().value('sidecar', 'channel_leases')
        self.monitor_leases = MonitorLeases(
            message_bus, self.reconnectable, channel, cycle,
            step=step, wait=wait, verbose=verbose)

    def on_channel(self, channel, *args):
        if channel == back_channel:
            self.monitor_leases.on_back_channel(*args)
        else:
            logger.warning("received data {} on unexpected channel {}".format(channel))

    # xxx no way to select/disable the 2 components (nodes and leases) for now
    async def run(self):
        logger.info("Starting monitor on {} nodes - report_wlan={}"
                    .format(len(self.monitor_nodes), self.report_wlan))
        # run n+1 tasks in parallel
        # one for leases,
        # plus one per node
        return asyncio.gather(
            self.monitor_leases.run_forever(),
            *[monitor_node.probe_forever(self.cycle,
                                         ping_timeout = self.ping_timeout,
                                         ssh_timeout = self.ssh_timeout)
              for monitor_node in self.monitor_nodes]
        )

    async def log(self):
        previous = 0
        while True:
            line = "".join([one_char_summary(mnode.info) for mnode in self.monitor_nodes])
            current = self.reconnectable.get_counter(self.main_channel)
            delta = "+ {}".format(current-previous)
            line += " {} emits ({})".format(current, delta)
            previous = current
            logger.info(line)
            await asyncio.sleep(self.log_period)
            

if __name__ == '__main__':
    import sys
    from node import Node
    # rebootnames = sys.argv[1:]
    message_bus = asyncio.Queue()
    # cycle = Config().value('monitor', 'cycle')
    # monitor = Monitor(rebootnames, message_bus, cycle=2, report_wlan=True)

    url = Config().value('sidecar', 'url')
    reconnectable = ReconnectableSocketIOMonitor(None, url, verbose=True)
    monitor = MonitorLeases(message_bus, reconnectable = reconnectable,
                            channel='info:leases',
                            cycle=10, step=1, wait=.1, verbose=True)
    loop = asyncio.get_event_loop()
    loop.run_until_complete(asyncio.gather(monitor.run_forever()))



#!/usr/bin/env python3
import sys
import random
import asyncio
import asyncssh

debug = False
#debug = True

# this class is specialized (see clientsession_closure below)
# this is how we can access self.node which is a reference to
# the corresponding Node object
class MySSHClientSession(asyncssh.SSHClientSession):
    def __init__(self, *args, **kwds):
        self.data = ""
        super().__init__(*args, **kwds)

    def data_received(self, data, datatype):
        # not adding a \n since it's already in there
        if debug: print('SSS DR: {}:{}-> {} [[of type {}]]'.
                        format(self.node, self.command, data, datatype), end='')
        self.data += data

    def connection_made(self, conn):
        if debug: print('SSS CM: {} {}'.format(self.node, conn))
        pass

    def connection_lost(self, exc):
        if exc:
            if debug: print('SSS CL: {} - exc={}'.format(self.node, exc))
        pass

    def eof_received(self):
        if debug: print('SSS EOF: {}'.format(self.node, self.command))


class MySSHClient(asyncssh.SSHClient):
    def connection_made(self, conn):
        if debug: print('SSC Connection made to %s.' % conn.get_extra_info('peername')[0])

    def auth_completed(self):
        if debug: print('SSC Authentication successful.')

class SshProxy:
    """
    talk to a Node's control interface using ssh
    """
    def __init__(self, node, username='root', verbose=False):
        self.node = node
        self.username = username
        self.verbose = verbose
        #
        self.hostname = self.node.control_hostname()
        self.status = None
        self.conn, self.client = None, None

    def __repr__(self):
        return "SshProxy {}".format(self.node)
    
    # make this an asynchroneous context manager
    # async with SshProxy(...) as ssh:
    #    
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # xxx this might be a little harsh, in the case
        # where an exception did occur
        await self.close()

    async def connect(self, timeout=None):
        try:
            # for private keys
            # also pass here client_keys = [some_list]
            # see http://asyncssh.readthedocs.org/en/latest/api.html#specifyingprivatekeys
            self.conn, self.client = await asyncio.wait_for(
                asyncssh.create_connection(
                    MySSHClient, self.hostname, username=self.username, known_hosts=None
                ),
                timeout = timeout)
            return True
        except (OSError, asyncssh.Error, asyncio.TimeoutError) as e:
            #await self.node.feedback('ssh_status', 'connect failed')
            #print('SshProxy.connect failed: {}'.format(e))
            self.conn, self.client = None, None
            return False

    async def run(self, command):
        """
        Run a command
        """
        class clientsession_closure(MySSHClientSession):
            def __init__(ssh_client_session, *args, **kwds):
                ssh_client_session.node = self.node
                ssh_client_session.command = command
                super().__init__(*args, **kwds)

        #print(5*'-', "running on ", self.hostname, ':', command)
        try:
            chan, session = await self.conn.create_session(clientsession_closure, command)
            await chan.wait_closed()
            return session.data
        except:
            return

    # >>> asyncio.iscoroutine(asyncssh.SSHClientConnection.close)
    # False
    async def close(self):
        if self.conn is not None:
            self.conn.close()
            await self.conn.wait_closed()
        self.conn = None

    async def wait_for(self, backoff, timeout=1.):
        """
        Wait until the ssh service is usable 
        """
        self.status = False
        while True:
            if self.verbose:
                await self.node.feedback('ssh_status', "trying to connect")
            self.status = await self.connect(timeout)
            if self.status:
                if self.verbose:
                    await self.node.feedback('ssh_status', "connection OK")
                await self.close()
                return self.status
            # random.random() is between 0. and 1. so we need something between 0.5 and 1.5
            random_backoff = (0.5 + random.random()) * backoff
            if self.verbose:
                await self.node.feedback(
                    'ssh_status',
                    "cannot connect, backing off for {:.3}s".format(random_backoff))
            await asyncio.sleep(random_backoff)


# mostly test-oriented
if __name__ == '__main__':        

    from rhubarbe.node import Node

    async def probe(h, message_bus):
        node = Node(h, message_bus)
        proxy = SshProxy(node, verbose=True)
        c = await proxy.connect()
        if not c:
            return False
        out1 = await proxy.run('cat /etc/lsb-release /etc/fedora-release 2> /dev/null')
        print("command1 returned {}".format(out1))
        out2 = await proxy.run('hostname')
        print("command2 returned {}".format(out2))
        await proxy.close()
        return True

    message_bus = asyncio.Queue()

    nodes = sys.argv[1:]
    tasks = [probe(node, message_bus) for node in nodes]

    retcods = asyncio.get_event_loop().run_until_complete(asyncio.gather(*tasks))

    for node, retcod in zip(nodes, retcods):
        print("{}:{}".format(node, retcod))



# from https://bitbucket.org/ssc/aiomas/src/92ddbe7f8c6dc86c03dd9d00c155a709ef330a89/aiomas/util.py?at=default&fileviewer=file-view-default

import asyncio
    
def self_manage(coro_or_future, ignore_cancel=True, loop=None):
    """Run :func:`asyncio.async()` with *coro_or_future* and set a callback
    that instantly raises all exceptions.

    If *ignore_cancel* is left ``True``, no exception is raised if the task was
    canceled.  If you also want to raise the ``CancelledError``, set the flag
    to ``False.``.

    Return an :class:`asyncio.Task` object.

    The difference between this function and :func:`asyncio.async()` subtle,
    but important if an error is raised by the task:

    :func:`asyncio.async()` returns a future (:class:`asyncio.Task` is
    a subclass of :class:`asyncio.Future`) for the task that you created.  By
    the time that future goes out of scope, asyncio checks if someone was
    interested in its result or not.  If the result was never retrieved, the
    exception is printed to *stderr*.

    If you call it like ``asyncio.async(mytask())`` (note that we don't keep
    a reference to the future here), an exception in *mytask* will pre printed
    immediately when the task is done.  If, however, we store a reference to
    the future (``fut = asyncio.async(mytask())``), the exception only gets
    printed when ``fut`` goes out of scope.  That means if, for example, an
    :class:`~aiomas.agent.Agent` creates a task and stores it as an instance
    attribute, our system may keep running for a long time after the exception
    has occured (or even block forever) and we won't see any stacktrace.  This
    is because the reference to the task is still there and we could, in
    theory, still retrieve the exception from there.

    Since this can make debugging very hard, this method simply registers a
    callback to the future.  The callback will try to get the result from the
    future when it is done and will thus print any exceptions immediately.

    """
    task = asyncio.ensure_future(coro_or_future, loop=loop)

    def cb(f):
        if f.cancelled() and ignore_cancel:
            return
        f.result()  # Let the future raise the exception

    task.add_done_callback(cb)

    return task



import json

from rhubarbe.singleton import Singleton
from rhubarbe.config import Config

class InventoryPhones(metaclass = Singleton):
    """
    A class for loading and storing the phones specifications
    typically from /etc/rhubarbe/inventory-phones.json
    """

    def __init__(self):
        the_config = Config()
        try:
            with open(the_config.value('testbed', 'inventory_phones_path')) as feed:
                self._phones = json.load(feed)
        except FileNotFoundError:
            self._phones = []

    def all_phones(self):
        """
        For now just return the json contents as is
        """
        return self._phones



#!/usr/bin/env python3

from __future__ import print_function

import sys
import os
import os.path
from distutils.core import setup
from rhubarbe.version import version as rhubarbe_version

# check python version
from sys import version_info
major, minor= version_info[0:2]
if not (major == 3 and minor >= 4):
    print("python 3.4 or higher is required")
    exit(1)

# read licence info
with open("README.md") as f:
    long_description = f.read()

# avoid this because it may be used early and the
# requirements are not met yet
#from rhubarbe.main import supported_subcommands
#cut'n'paste from the rhubarbe help message
rhubarbe_help = "nodes,status,on,off,reset,info,usrpstatus,usrpon,usrpoff,load,save,wait,monitor,monitorphones,accounts,leases,images,resolve,share,inventory,config,version"
supported_subcommands = rhubarbe_help.split(",")
subcommand_symlinks = [ "bin/rhubarbe-{}".format(subcommand) for subcommand in supported_subcommands ]
if sys.argv[1] in ('sdist'):
    for subcommand in supported_subcommands:
        link = "bin/rhubarbe-{}".format(subcommand)
        if not os.path.exists(link):
            os.symlink("rhubarbe", link)

### requirements - used by pip install
# *NOTE* for ubuntu: also run this beforehand
# apt-get -y install libffi-dev
# which is required before pip can install asyncssh
required_modules = [
    # version 1.0 breaks our code
    'telnetlib3==0.5.0',
    'aiohttp',
    'asyncssh',
    'progressbar33',
    # for monitor
    'socketIO-client',
    'asynciojobs',
    'aioxmlrpc',
]

setup(
    name             = "rhubarbe",
    version          = rhubarbe_version,
    description      = "Testbed Management Framework for R2Lab",
    long_description = long_description,
    license          = license,
    author           = "Thierry Parmentelat",
    author_email     = "thierry.parmentelat@inria.fr",
    download_url     = "http://github/build.onelab.eu/rhubarbe/rhubarbe-{v}.tar.gz".format(v=rhubarbe_version),
    url              = "https://github.com/parmentelat/rhubarbe/tree/master",
    platforms        = "Linux",
    packages         = [ 'rhubarbe' ],
    data_files       =
      [ ('/etc/rhubarbe', [
          'rhubarbe.conf',
          'inventory-nodes.json.template',
          'inventory-phones.json.template',
      ] ) ],
    scripts          = [
        'bin/rhubarbe',
    ] + subcommand_symlinks,
    install_requires = required_modules,
)
